package java.time;

import java.time.format.SignStyle;
import java.time.format.DateTimeFormatterBuilder;
import java.io.DataInput;
import java.io.IOException;
import java.io.DataOutput;
import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.time.temporal.TemporalQueries;
import java.time.temporal.TemporalQuery;
import java.time.temporal.TemporalAmount;
import java.time.temporal.UnsupportedTemporalTypeException;
import java.time.temporal.ValueRange;
import java.time.temporal.ChronoUnit;
import java.time.temporal.TemporalUnit;
import java.time.temporal.TemporalField;
import java.time.chrono.Chronology;
import java.time.chrono.IsoChronology;
import java.time.temporal.TemporalAccessor;
import java.time.temporal.ChronoField;
import java.util.Objects;
import java.time.format.DateTimeFormatter;
import java.io.Serializable;
import java.time.temporal.TemporalAdjuster;
import java.time.temporal.Temporal;

public final class YearMonth implements Temporal, TemporalAdjuster, Comparable<YearMonth>, Serializable
{
    private static final long serialVersionUID = 4183400860270640070L;
    private static final DateTimeFormatter PARSER;
    private final int year;
    private final int month;
    
    public static YearMonth now() {
        return now(Clock.systemDefaultZone());
    }
    
    public static YearMonth now(ZoneId zoneId) {
        return now(Clock.system(zoneId));
    }
    
    public static YearMonth now(Clock clock) {
        LocalDate now = LocalDate.now(clock);
        return of(now.getYear(), now.getMonth());
    }
    
    public static YearMonth of(int n, Month obj) {
        Objects.requireNonNull(obj, "month");
        return of(n, obj.getValue());
    }
    
    public static YearMonth of(int n, int n2) {
        ChronoField.YEAR.checkValidValue((long)n);
        ChronoField.MONTH_OF_YEAR.checkValidValue((long)n2);
        return new YearMonth(n, n2);
    }
    
    public static YearMonth from(TemporalAccessor from) {
        if ((from instanceof YearMonth)) {
            return (YearMonth)from;
        }
        Objects.requireNonNull(from, "temporal");
        try {
            if (!IsoChronology.INSTANCE.equals(Chronology.from(from))) {
                from = LocalDate.from(from);
            }
            return of(from.get(ChronoField.YEAR), from.get(ChronoField.MONTH_OF_YEAR));
        }
        catch (DateTimeException ex) {
            throw new DateTimeException("Unable to obtain YearMonth from TemporalAccessor: " + from + " of type " + from.getClass().getName(), ex);
        }
    }
    
    public static YearMonth parse(CharSequence charSequence) {
        return parse(charSequence, YearMonth.PARSER);
    }
    
    public static YearMonth parse(CharSequence text, DateTimeFormatter obj) {
        Objects.requireNonNull(obj, "formatter");
        return (YearMonth)obj.parse(text, YearMonth::from);
    }
    
    private YearMonth(int year, int month) {
        this.year = year;
        this.month = month;
    }
    
    private YearMonth with(int n, int n2) {
        if (this.year == n && this.month == n2) {
            return this;
        }
        return new YearMonth(n, n2);
    }
    
    @Override
    public boolean isSupported(TemporalField temporalField) {
        if ((temporalField instanceof ChronoField)) {
            return temporalField == ChronoField.YEAR || temporalField == ChronoField.MONTH_OF_YEAR || temporalField == ChronoField.PROLEPTIC_MONTH || temporalField == ChronoField.YEAR_OF_ERA || temporalField == ChronoField.ERA;
        }
        return temporalField != null && temporalField.isSupportedBy(this);
    }
    
    @Override
    public boolean isSupported(TemporalUnit temporalUnit) {
        if ((temporalUnit instanceof ChronoUnit)) {
            return temporalUnit == ChronoUnit.MONTHS || temporalUnit == ChronoUnit.YEARS || temporalUnit == ChronoUnit.DECADES || temporalUnit == ChronoUnit.CENTURIES || temporalUnit == ChronoUnit.MILLENNIA || temporalUnit == ChronoUnit.ERAS;
        }
        return temporalUnit != null && temporalUnit.isSupportedBy(this);
    }
    
    @Override
    public ValueRange range(TemporalField field) {
        if (field == ChronoField.YEAR_OF_ERA) {
            return (this.getYear() <= 0) ? ValueRange.of(1L, 1000000000L) : ValueRange.of(1L, 999999999L);
        }
        return super.range(field);
    }
    
    @Override
    public int get(TemporalField field) {
        return this.range(field).checkValidIntValue(this.getLong(field), field);
    }
    
    @Override
    public long getLong(TemporalField obj) {
        if ((obj instanceof ChronoField)) {
            switch ((ChronoField)obj) {
                case MONTH_OF_YEAR: {
                    return (long)this.month;
                }
                case PROLEPTIC_MONTH: {
                    return this.getProlepticMonth();
                }
                case YEAR_OF_ERA: {
                    return (long)((this.year < 1) ? (1 - this.year) : this.year);
                }
                case YEAR: {
                    return (long)this.year;
                }
                case ERA: {
                    return (long)(this.year >= 1);
                }
                default: {
                    throw new UnsupportedTemporalTypeException("Unsupported field: " + obj);
                }
            }
        }
        else {
            return obj.getFrom(this);
        }
    }
    
    private long getProlepticMonth() {
        return (long)this.year * 12L + (long)this.month - 1L;
    }
    
    public int getYear() {
        return this.year;
    }
    
    public int getMonthValue() {
        return this.month;
    }
    
    public Month getMonth() {
        return Month.of(this.month);
    }
    
    public boolean isLeapYear() {
        return IsoChronology.INSTANCE.isLeapYear((long)this.year);
    }
    
    public boolean isValidDay(int n) {
        return n >= 1 && n <= this.lengthOfMonth();
    }
    
    public int lengthOfMonth() {
        return this.getMonth().length(this.isLeapYear());
    }
    
    public int lengthOfYear() {
        return this.isLeapYear() ? 366 : 365;
    }
    
    @Override
    public YearMonth with(TemporalAdjuster temporalAdjuster) {
        return (YearMonth)temporalAdjuster.adjustInto(this);
    }
    
    @Override
    public YearMonth with(TemporalField obj, long value) {
        if ((obj instanceof ChronoField)) {
            ChronoField chronoField = (ChronoField)obj;
            chronoField.checkValidValue(value);
            switch (chronoField) {
                case MONTH_OF_YEAR: {
                    return this.withMonth((int)value);
                }
                case PROLEPTIC_MONTH: {
                    return this.plusMonths(value - this.getProlepticMonth());
                }
                case YEAR_OF_ERA: {
                    return this.withYear((int)((this.year < 1) ? (1L - value) : value));
                }
                case YEAR: {
                    return this.withYear((int)value);
                }
                case ERA: {
                    return (this.getLong(ChronoField.ERA) == value) ? this : this.withYear(1 - this.year);
                }
                default: {
                    throw new UnsupportedTemporalTypeException("Unsupported field: " + obj);
                }
            }
        }
        else {
            return (YearMonth)obj.adjustInto(this, value);
        }
    }
    
    public YearMonth withYear(int n) {
        ChronoField.YEAR.checkValidValue((long)n);
        return this.with(n, this.month);
    }
    
    public YearMonth withMonth(int n) {
        ChronoField.MONTH_OF_YEAR.checkValidValue((long)n);
        return this.with(this.year, n);
    }
    
    @Override
    public YearMonth plus(TemporalAmount temporalAmount) {
        return (YearMonth)temporalAmount.addTo(this);
    }
    
    @Override
    public YearMonth plus(long n, TemporalUnit obj) {
        if ((obj instanceof ChronoUnit)) {
            switch ((ChronoUnit)obj) {
                case MONTHS: {
                    return this.plusMonths(n);
                }
                case YEARS: {
                    return this.plusYears(n);
                }
                case DECADES: {
                    return this.plusYears(Math.multiplyExact(n, 10L));
                }
                case CENTURIES: {
                    return this.plusYears(Math.multiplyExact(n, 100L));
                }
                case MILLENNIA: {
                    return this.plusYears(Math.multiplyExact(n, 1000L));
                }
                case ERAS: {
                    return this.with(ChronoField.ERA, Math.addExact(this.getLong(ChronoField.ERA), n));
                }
                default: {
                    throw new UnsupportedTemporalTypeException("Unsupported unit: " + obj);
                }
            }
        }
        else {
            return (YearMonth)obj.addTo(this, n);
        }
    }
    
    public YearMonth plusYears(long n) {
        if (n == 0L) {
            return this;
        }
        return this.with(ChronoField.YEAR.checkValidIntValue((long)this.year + n), this.month);
    }
    
    public YearMonth plusMonths(long n) {
        if (n == 0L) {
            return this;
        }
        long n2 = (long)this.year * 12L + (long)(this.month - 1) + n;
        return this.with(ChronoField.YEAR.checkValidIntValue(Math.floorDiv(n2, 12L)), (int)Math.floorMod(n2, 12L) + 1);
    }
    
    @Override
    public YearMonth minus(TemporalAmount temporalAmount) {
        return (YearMonth)temporalAmount.subtractFrom(this);
    }
    
    @Override
    public YearMonth minus(long n, TemporalUnit temporalUnit) {
        return (n == Long.MIN_VALUE) ? this.plus(Long.MAX_VALUE, temporalUnit).plus(1L, temporalUnit) : this.plus(-n, temporalUnit);
    }
    
    public YearMonth minusYears(long n) {
        return (n == Long.MIN_VALUE) ? this.plusYears(Long.MAX_VALUE).plusYears(1L) : this.plusYears(-n);
    }
    
    public YearMonth minusMonths(long n) {
        return (n == Long.MIN_VALUE) ? this.plusMonths(Long.MAX_VALUE).plusMonths(1L) : this.plusMonths(-n);
    }
    
    @Override
    public <R> R query(TemporalQuery<R> query) {
        if (query == TemporalQueries.chronology()) {
            return IsoChronology.INSTANCE;
        }
        if (query == TemporalQueries.precision()) {
            return ChronoUnit.MONTHS;
        }
        return super.query(query);
    }
    
    @Override
    public Temporal adjustInto(Temporal temporal) {
        if (!Chronology.from(temporal).equals(IsoChronology.INSTANCE)) {
            throw new DateTimeException("Adjustment only supported on ISO date-time");
        }
        return temporal.with(ChronoField.PROLEPTIC_MONTH, this.getProlepticMonth());
    }
    
    @Override
    public long until(Temporal temporal, TemporalUnit obj) {
        YearMonth from = from(temporal);
        if ((obj instanceof ChronoUnit)) {
            long n = from.getProlepticMonth() - this.getProlepticMonth();
            switch ((ChronoUnit)obj) {
                case MONTHS: {
                    return n;
                }
                case YEARS: {
                    return n / 12L;
                }
                case DECADES: {
                    return n / 120L;
                }
                case CENTURIES: {
                    return n / 1200L;
                }
                case MILLENNIA: {
                    return n / 12000L;
                }
                case ERAS: {
                    return from.getLong(ChronoField.ERA) - this.getLong(ChronoField.ERA);
                }
                default: {
                    throw new UnsupportedTemporalTypeException("Unsupported unit: " + obj);
                }
            }
        }
        else {
            return obj.between(this, from);
        }
    }
    
    public String format(DateTimeFormatter obj) {
        Objects.requireNonNull(obj, "formatter");
        return obj.format(this);
    }
    
    public LocalDate atDay(int n) {
        return LocalDate.of(this.year, this.month, n);
    }
    
    public LocalDate atEndOfMonth() {
        return LocalDate.of(this.year, this.month, this.lengthOfMonth());
    }
    
    @Override
    public int compareTo(YearMonth yearMonth) {
        int n = this.year - yearMonth.year;
        if (n == 0) {
            n = this.month - yearMonth.month;
        }
        return n;
    }
    
    public boolean isAfter(YearMonth yearMonth) {
        return this.compareTo(yearMonth) > 0;
    }
    
    public boolean isBefore(YearMonth yearMonth) {
        return this.compareTo(yearMonth) < 0;
    }
    
    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if ((o instanceof YearMonth)) {
            YearMonth yearMonth = (YearMonth)o;
            return this.year == yearMonth.year && this.month == yearMonth.month;
        }
        return false;
    }
    
    @Override
    public int hashCode() {
        return this.year ^ this.month << 27;
    }
    
    @Override
    public String toString() {
        int abs = Math.abs(this.year);
        StringBuilder sb = new StringBuilder(9);
        if (abs < 1000) {
            if (this.year < 0) {
                sb.append(this.year - 10000).deleteCharAt(1);
            }
            else {
                sb.append(this.year + 10000).deleteCharAt(0);
            }
        }
        else {
            sb.append(this.year);
        }
        return sb.append((this.month < 10) ? "-0" : "-").append(this.month).toString();
    }
    
    private Object writeReplace() {
        return new Ser((byte)12, this);
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws InvalidObjectException {
        throw new InvalidObjectException("Deserialization via serialization delegate");
    }
    
    void writeExternal(DataOutput dataOutput) throws IOException {
        dataOutput.writeInt(this.year);
        dataOutput.writeByte(this.month);
    }
    
    static YearMonth readExternal(DataInput dataInput) throws IOException {
        return of(dataInput.readInt(), (int)dataInput.readByte());
    }
    
    static {
        PARSER = new DateTimeFormatterBuilder().appendValue(ChronoField.YEAR, 4, 10, SignStyle.EXCEEDS_PAD).appendLiteral('-').appendValue(ChronoField.MONTH_OF_YEAR, 2).toFormatter();
    }
}
